"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const events_1 = require("events");
const msgpack_lite_1 = require("msgpack-lite");
function Response(encoder, request_id) {
    this.encodeStream = encoder;
    this._request_id = request_id;
}
Response.prototype.send = function (resp, is_error) {
    if (this._sent) {
        throw new Error('Response to id ' + this._request_id + ' already sent');
    }
    if (is_error) {
        this.encodeStream.write([1, this._request_id, resp, null]);
    }
    else {
        this.encodeStream.write([1, this._request_id, null, resp]);
    }
    this._sent = true;
};
class Session extends events_1.EventEmitter {
    constructor(types) {
        super();
        this.pending = new Map();
        this.nextRequestId = 1;
        this.addTypes(types);
        this.encodeStream = msgpack_lite_1.default.createEncodeStream();
        this.decodeStream = msgpack_lite_1.default.createDecodeStream();
        this.decodeStream.on('data', (msg) => {
            this._parse_message(msg);
        });
        this.decodeStream.on('end', () => {
            this.detach();
            this.emit('detach');
        });
    }
    addTypes(types) {
        if (types)
            for (var i = 0, l = types.length; i < l; i++) {
                var type = types[i];
                this._msgpack.register(type.code, type.constructor, type.encode, type.decode);
            }
    }
    attach(writer, reader) {
        this.encoderStream.pipe(writer);
        reader.pipe(this.decodeStream);
        this.writer = writer;
        this.reader = reader;
    }
    detach() {
        this.encodeStream.unpipe(this.writer);
        this.reader.unpipe(this.decodeStream);
    }
    request(method, args, cb) {
        this.nextRequestId = this.nextRequestId + 1;
        this.encodeStream.write([0, this.nextRequestId, method, args]);
        this.pending.set(this.nextRequestId, cb);
    }
    notify(method, args) {
        this.encodeStream.write([2, method, args]);
    }
    _parse_message(msg) {
        var msg_type = msg[0];
        if (msg_type === 0) {
            // request
            //   - msg[1]: id
            //   - msg[2]: method name
            //   - msg[3]: arguments
            this.emit('request', msg[2].toString(), msg[3], new Response(this.encodeStream, msg[1]));
        }
        else if (msg_type === 1) {
            // response to a previous request:
            //   - msg[1]: the id
            //   - msg[2]: error(if any)
            //   - msg[3]: result(if not errored)
            var id = msg[1];
            var handler = this.pending.get(id);
            this.pending.remove(id);
            handler(msg[2], msg[3]);
        }
        else if (msg_type === 2) {
            // notification/event
            //   - msg[1]: event name
            //   - msg[2]: arguments
            this.emit('notification', msg[1].toString(), msg[2]);
        }
        else {
            this.encodeStream.write([1, 0, 'Invalid message type', null]);
        }
    }
}
exports.Session = Session;
